/*

	Paint in-memory DIB into HDC

*/

#include "bullets.h"

/*	PAINTDIB  --  Paint a DIB into an HDC in an aesthetic
				  manner.  The rectangle into which the
				  (possibly stretched) bitmap was written is
				  returned in bmprect. */

BOOL paintDIB(HWND hwnd, HDC hdc, LPBYTE bitmap, LPRECT bmprect)
{
	int bx, by;
	RECT cr;
	BITMAPFILEHEADER FAR *bfh;
	BITMAPINFOHEADER FAR *bmi;
	char *bits;
	HPALETTE bpal = NULL, opal;
    int i;
    LPLOGPALETTE lp;
	HBITMAP ibmap = NULL;
    BITMAPINFO FAR *bh;
    unsigned short FAR *palidx;
    LPSTR lpalette;
    unsigned short FAR *savepal;
	
	savepal = (unsigned short FAR *) GlobalAlloc(GPTR,
				sizeof(LOGPALETTE) +
				sizeof(PALETTEENTRY) * 256 +
				sizeof(short) * 256);
	if (savepal == NULL) {
		return FALSE;
	}
	lpalette = ((LPSTR) savepal) + 256 * sizeof(short); 
	GetClientRect(hwnd, &cr);
    
    bfh = (BITMAPFILEHEADER FAR *) bitmap;
    bmi = (BITMAPINFOHEADER FAR *) (bitmap + sizeof(BITMAPFILEHEADER));
    bh = (BITMAPINFO FAR *) bmi; 
    bx = (int) bmi->biWidth;
    by = (int) bmi->biHeight;
    
    if ((bmi->biClrUsed == 0) && (bmi->biBitCount < 24)) {
        bmi->biClrUsed = 1L << bmi->biBitCount;
    }

	bits = (char FAR *) (bitmap + bfh->bfOffBits);

	if (bmi->biBitCount < 24) {
		lp = (LOGPALETTE FAR *) lpalette;
		palidx = (unsigned short FAR *) bh->bmiColors;
		memcpy(savepal, palidx, 256 * sizeof(short));
		lp->palVersion = 0x0300;
		lp->palNumEntries = (WORD) bmi->biClrUsed;
		for (i = 0; i < ((int) bmi->biClrUsed); i++) {
			lp->palPalEntry[i].peRed = bh->bmiColors[i].rgbRed;
			lp->palPalEntry[i].peGreen = bh->bmiColors[i].rgbGreen;
			lp->palPalEntry[i].peBlue = bh->bmiColors[i].rgbBlue;
			lp->palPalEntry[i].peFlags = PC_NOCOLLAPSE;
		}
		bpal = CreatePalette(lp);
		if (bpal != NULL) {
			opal = SelectPalette(hdc, bpal, FALSE);
			RealizePalette(hdc);
		}
	}

	ibmap = CreateDIBitmap(hdc, bmi, CBM_INIT,
	            bits, (BITMAPINFO FAR *) bmi, DIB_RGB_COLORS);
	if (bpal != NULL) {
		memcpy(palidx, savepal, 256 * sizeof(short));
	}
	if (ibmap != NULL) {
		HDC hMemDC;
		HBITMAP obmap;
		HPALETTE mopal;

		hMemDC = CreateCompatibleDC(hdc);
		if (bpal != NULL) {
			mopal = SelectPalette(hMemDC, bpal, FALSE);
		}
		obmap = SelectObject(hMemDC, ibmap);
		if (bx <= cr.right && by <= cr.bottom) {
			BitBlt(hdc, (cr.right - bx) / 2, (cr.bottom - by) / 2,
			    bx, by, hMemDC, 0, 0, SRCCOPY);
			bmprect->left = (cr.right - bx) / 2;
			bmprect->top = (cr.bottom - by) / 2;
			bmprect->right = bmprect->left + bx;
			bmprect->bottom = bmprect->top + by;
		} else {
			int nx, ny;
			double xshrink = ((double) bx) / cr.right,
				   yshrink = ((double) by) / cr.bottom;

			if (xshrink > yshrink) {
				nx = cr.right;
				ny = (int) (by / xshrink);
			} else {
				ny = cr.bottom;
				nx = (int) (bx / yshrink);
			}
			SetStretchBltMode(hdc, STRETCH_DELETESCANS);
			StretchBlt(hdc, (cr.right - nx) / 2, (cr.bottom - ny) / 2, nx, ny,
				hMemDC, 0, 0, bx, by, SRCCOPY);
			bmprect->left = (cr.right - nx) / 2;
			bmprect->top = (cr.bottom - ny) / 2;
			bmprect->right = bmprect->left + nx;
			bmprect->bottom = bmprect->top + ny;
		}
		SelectObject(hMemDC, obmap);
		DeleteObject(ibmap);
		if (bpal != NULL) {
			SelectPalette(hdc, opal, FALSE);
			SelectPalette(hMemDC, mopal, FALSE);
			DeleteObject(bpal);
		}
		DeleteDC(hMemDC);
	}
	GlobalFree(savepal);
	return TRUE;
}
